@@ -5,7 +5,7 @@
  * version 2.0 (the "License"); you may not use this file except in compliance
  * with the License. You may obtain a copy of the License at:
  *
- *   http://www.apache.org/licenses/LICENSE-2.0
+ *   https://www.apache.org/licenses/LICENSE-2.0
  *
  * Unless required by applicable law or agreed to in writing, software
  * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
@@ -15,10 +15,13 @@
  */
 package io.netty.buffer;
 
+import io.netty.util.ByteProcessor;
 import io.netty.util.ReferenceCountUtil;
+import io.netty.util.internal.ObjectUtil;
 import io.netty.util.internal.PlatformDependent;
-import org.junit.Assume;
-import org.junit.Test;
+import org.junit.jupiter.api.Assumptions;
+import org.junit.jupiter.api.Test;
+import org.junit.jupiter.api.function.Executable;
 
 import java.nio.ByteBuffer;
 import java.nio.ByteOrder;
@@ -35,17 +38,16 @@ import static io.netty.buffer.Unpooled.compositeBuffer;
 import static io.netty.buffer.Unpooled.directBuffer;
 import static io.netty.buffer.Unpooled.wrappedBuffer;
 import static io.netty.util.internal.EmptyArrays.EMPTY_BYTES;
-import static org.hamcrest.CoreMatchers.instanceOf;
-import static org.hamcrest.CoreMatchers.is;
-import static org.junit.Assert.assertArrayEquals;
-import static org.junit.Assert.assertEquals;
-import static org.junit.Assert.assertFalse;
-import static org.junit.Assert.assertNotNull;
-import static org.junit.Assert.assertNotSame;
-import static org.junit.Assert.assertSame;
-import static org.junit.Assert.assertThat;
-import static org.junit.Assert.assertTrue;
-import static org.junit.Assert.fail;
+import static org.junit.jupiter.api.Assertions.assertArrayEquals;
+import static org.junit.jupiter.api.Assertions.assertEquals;
+import static org.junit.jupiter.api.Assertions.assertFalse;
+import static org.junit.jupiter.api.Assertions.assertInstanceOf;
+import static org.junit.jupiter.api.Assertions.assertNotNull;
+import static org.junit.jupiter.api.Assertions.assertNotSame;
+import static org.junit.jupiter.api.Assertions.assertSame;
+import static org.junit.jupiter.api.Assertions.assertThrows;
+import static org.junit.jupiter.api.Assertions.assertTrue;
+import static org.junit.jupiter.api.Assertions.fail;
 
 /**
  * An abstract test class for composite channel buffers
@@ -57,15 +59,12 @@ public abstract class AbstractCompositeByteBufTest extends AbstractByteBufTest {
     private final ByteOrder order;
 
     protected AbstractCompositeByteBufTest(ByteOrder order) {
-        if (order == null) {
-            throw new NullPointerException("order");
-        }
-        this.order = order;
+        this.order = ObjectUtil.checkNotNull(order, "order");
     }
 
     @Override
     protected ByteBuf newBuffer(int length, int maxCapacity) {
-        Assume.assumeTrue(maxCapacity == Integer.MAX_VALUE);
+        Assumptions.assumeTrue(maxCapacity == Integer.MAX_VALUE);
 
         List<ByteBuf> buffers = new ArrayList<ByteBuf>();
         for (int i = 0; i < length + 45; i += 45) {
@@ -90,7 +89,20 @@ public abstract class AbstractCompositeByteBufTest extends AbstractByteBufTest {
             buffers.add(EMPTY_BUFFER);
         }
 
-        ByteBuf buffer = wrappedBuffer(Integer.MAX_VALUE, buffers.toArray(new ByteBuf[0])).order(order);
+        ByteBuf buffer;
+        // Ensure that we are really testing a CompositeByteBuf
+        switch (buffers.size()) {
+            case 0:
+                buffer = compositeBuffer(Integer.MAX_VALUE);
+                break;
+            case 1:
+                buffer = compositeBuffer(Integer.MAX_VALUE).addComponent(buffers.get(0));
+                break;
+            default:
+                buffer = wrappedBuffer(Integer.MAX_VALUE, buffers.toArray(new ByteBuf[0]));
+                break;
+        }
+        buffer = buffer.order(order);
 
         // Truncate to the requested capacity.
         buffer.capacity(length);
@@ -102,6 +114,10 @@ public abstract class AbstractCompositeByteBufTest extends AbstractByteBufTest {
         return buffer;
     }
 
+    protected CompositeByteBuf newCompositeBuffer() {
+        return compositeBuffer();
+    }
+
     // Composite buffer does not waste bandwidth on discardReadBytes, but
     // the test will fail in strict mode.
     @Override
@@ -109,6 +125,13 @@ public abstract class AbstractCompositeByteBufTest extends AbstractByteBufTest {
         return false;
     }
 
+    @Test
+    public void testIsContiguous() {
+        ByteBuf buf = newBuffer(4);
+        assertFalse(buf.isContiguous());
+        buf.release();
+    }
+
     /**
      * Tests the "getBufferFor" method
      */
@@ -128,8 +151,8 @@ public abstract class AbstractCompositeByteBufTest extends AbstractByteBufTest {
             ByteBuf _buf = buf.componentAtOffset(index++);
             assertNotNull(_buf);
             assertTrue(_buf.capacity() > 0);
-            assertNotNull(_buf.getByte(0));
-            assertNotNull(_buf.getByte(_buf.readableBytes() - 1));
+            assertTrue(_buf.getByte(0) > 0);
+            assertTrue(_buf.getByte(_buf.readableBytes() - 1) > 0);
         }
 
         buf.release();
@@ -626,10 +649,10 @@ public abstract class AbstractCompositeByteBufTest extends AbstractByteBufTest {
     public void testComponentMustBeDuplicate() {
         CompositeByteBuf buf = compositeBuffer();
         buf.addComponent(buffer(4, 6).setIndex(1, 3));
-        assertThat(buf.component(0), is(instanceOf(AbstractDerivedByteBuf.class)));
-        assertThat(buf.component(0).capacity(), is(4));
-        assertThat(buf.component(0).maxCapacity(), is(6));
-        assertThat(buf.component(0).readableBytes(), is(2));
+        assertInstanceOf(AbstractDerivedByteBuf.class, buf.component(0));
+        assertEquals(4, buf.component(0).capacity());
+        assertEquals(6, buf.component(0).maxCapacity());
+        assertEquals(2, buf.component(0).readableBytes());
         buf.release();
     }
 
@@ -640,19 +663,19 @@ public abstract class AbstractCompositeByteBufTest extends AbstractByteBufTest {
         ByteBuf c3 = buffer().writeByte(3).retain(2);
 
         CompositeByteBuf buf = compositeBuffer();
-        assertThat(buf.refCnt(), is(1));
+        assertEquals(1, buf.refCnt());
         buf.addComponents(c1, c2, c3);
 
-        assertThat(buf.refCnt(), is(1));
+        assertEquals(1, buf.refCnt());
 
         // Ensure that c[123]'s refCount did not change.
-        assertThat(c1.refCnt(), is(1));
-        assertThat(c2.refCnt(), is(2));
-        assertThat(c3.refCnt(), is(3));
+        assertEquals(1, c1.refCnt());
+        assertEquals(2, c2.refCnt());
+        assertEquals(3, c3.refCnt());
 
-        assertThat(buf.component(0).refCnt(), is(1));
-        assertThat(buf.component(1).refCnt(), is(2));
-        assertThat(buf.component(2).refCnt(), is(3));
+        assertEquals(1, buf.component(0).refCnt());
+        assertEquals(2, buf.component(1).refCnt());
+        assertEquals(3, buf.component(2).refCnt());
 
         c3.release(2);
         c2.release();
@@ -672,24 +695,24 @@ public abstract class AbstractCompositeByteBufTest extends AbstractByteBufTest {
         bufB.addComponents(bufA);
 
         // Ensure that bufA.refCnt() did not change.
-        assertThat(bufA.refCnt(), is(1));
+        assertEquals(1, bufA.refCnt());
 
         // Ensure that c[123]'s refCnt did not change.
-        assertThat(c1.refCnt(), is(1));
-        assertThat(c2.refCnt(), is(2));
-        assertThat(c3.refCnt(), is(3));
+        assertEquals(1, c1.refCnt());
+        assertEquals(2, c2.refCnt());
+        assertEquals(3, c3.refCnt());
 
         // This should decrease bufA.refCnt().
         bufB.release();
-        assertThat(bufB.refCnt(), is(0));
+        assertEquals(0, bufB.refCnt());
 
         // Ensure bufA.refCnt() changed.
-        assertThat(bufA.refCnt(), is(0));
+        assertEquals(0, bufA.refCnt());
 
         // Ensure that c[123]'s refCnt also changed due to the deallocation of bufA.
-        assertThat(c1.refCnt(), is(0));
-        assertThat(c2.refCnt(), is(1));
-        assertThat(c3.refCnt(), is(2));
+        assertEquals(0, c1.refCnt());
+        assertEquals(1, c2.refCnt());
+        assertEquals(2, c3.refCnt());
 
         c3.release(2);
         c2.release();
@@ -702,20 +725,20 @@ public abstract class AbstractCompositeByteBufTest extends AbstractByteBufTest {
         ByteBuf c3 = buffer().writeByte(3).retain(2);
 
         CompositeByteBuf buf = compositeBuffer();
-        assertThat(buf.refCnt(), is(1));
+        assertEquals(1, buf.refCnt());
 
         List<ByteBuf> components = new ArrayList<ByteBuf>();
         Collections.addAll(components, c1, c2, c3);
         buf.addComponents(components);
 
         // Ensure that c[123]'s refCount did not change.
-        assertThat(c1.refCnt(), is(1));
-        assertThat(c2.refCnt(), is(2));
-        assertThat(c3.refCnt(), is(3));
+        assertEquals(1, c1.refCnt());
+        assertEquals(2, c2.refCnt());
+        assertEquals(3, c3.refCnt());
 
-        assertThat(buf.component(0).refCnt(), is(1));
-        assertThat(buf.component(1).refCnt(), is(2));
-        assertThat(buf.component(2).refCnt(), is(3));
+        assertEquals(1, buf.component(0).refCnt());
+        assertEquals(2, buf.component(1).refCnt());
+        assertEquals(3, buf.component(2).refCnt());
 
         c3.release(2);
         c2.release();
@@ -731,11 +754,11 @@ public abstract class AbstractCompositeByteBufTest extends AbstractByteBufTest {
                         .addComponent(wrappedBuffer(new byte[]{3, 4})).slice(1, 2));
 
         ByteBuffer[] nioBuffers = buf.nioBuffers(0, 2);
-        assertThat(nioBuffers.length, is(2));
-        assertThat(nioBuffers[0].remaining(), is(1));
-        assertThat(nioBuffers[0].get(), is((byte) 2));
-        assertThat(nioBuffers[1].remaining(), is(1));
-        assertThat(nioBuffers[1].get(), is((byte) 3));
+        assertEquals(2, nioBuffers.length);
+        assertEquals(1, nioBuffers[0].remaining());
+        assertEquals(2, nioBuffers[0].get());
+        assertEquals(1, nioBuffers[1].remaining());
+        assertEquals(3, nioBuffers[1].get());
 
         buf.release();
     }
@@ -977,7 +1000,27 @@ public abstract class AbstractCompositeByteBufTest extends AbstractByteBufTest {
     @Override
     @Test
     public void testInternalNioBuffer() {
-        // ignore
+        CompositeByteBuf buf = compositeBuffer();
+        assertEquals(0, buf.internalNioBuffer(0, 0).remaining());
+
+        // If non-derived buffer is added, its internal buffer should be returned
+        ByteBuf concreteBuffer = directBuffer().writeByte(1);
+        buf.addComponent(concreteBuffer);
+        assertSame(concreteBuffer.internalNioBuffer(0, 1), buf.internalNioBuffer(0, 1));
+        buf.release();
+
+        // In derived cases, the original internal buffer must not be used
+        buf = compositeBuffer();
+        concreteBuffer = directBuffer().writeByte(1);
+        buf.addComponent(concreteBuffer.slice());
+        assertNotSame(concreteBuffer.internalNioBuffer(0, 1), buf.internalNioBuffer(0, 1));
+        buf.release();
+
+        buf = compositeBuffer();
+        concreteBuffer = directBuffer().writeByte(1);
+        buf.addComponent(concreteBuffer.duplicate());
+        assertNotSame(concreteBuffer.internalNioBuffer(0, 1), buf.internalNioBuffer(0, 1));
+        buf.release();
     }
 
     @Test
@@ -1098,8 +1141,17 @@ public abstract class AbstractCompositeByteBufTest extends AbstractByteBufTest {
 
     @Test
     public void testAddFlattenedComponents() {
+        testAddFlattenedComponents(false);
+    }
+
+    @Test
+    public void testAddFlattenedComponentsWithWrappedComposite() {
+        testAddFlattenedComponents(true);
+    }
+
+    private void testAddFlattenedComponents(boolean addWrapped) {
         ByteBuf b1 = Unpooled.wrappedBuffer(new byte[] { 1, 2, 3 });
-        CompositeByteBuf newComposite = Unpooled.compositeBuffer()
+        CompositeByteBuf newComposite = newCompositeBuffer()
                 .addComponent(true, b1)
                 .addFlattenedComponents(true, b1.retain())
                 .addFlattenedComponents(true, Unpooled.EMPTY_BUFFER);
@@ -1122,7 +1174,7 @@ public abstract class AbstractCompositeByteBufTest extends AbstractByteBufTest {
         ByteBuf s4 = s2.retainedSlice(0, 2);
         buffer.release();
 
-        ByteBuf compositeToAdd = Unpooled.compositeBuffer()
+        CompositeByteBuf compositeToAdd = compositeBuffer()
             .addComponent(s1)
             .addComponent(Unpooled.EMPTY_BUFFER)
             .addComponents(s2, s3, s4);
@@ -1135,6 +1187,9 @@ public abstract class AbstractCompositeByteBufTest extends AbstractByteBufTest {
 
         ByteBuf compositeCopy = compositeToAdd.copy();
 
+        if (addWrapped) {
+            compositeToAdd = new WrappedCompositeByteBuf(compositeToAdd);
+        }
         newComposite.addFlattenedComponents(true, compositeToAdd);
 
         // verify that added range matches
@@ -1163,7 +1218,7 @@ public abstract class AbstractCompositeByteBufTest extends AbstractByteBufTest {
 
     @Test
     public void testIterator() {
-        CompositeByteBuf cbuf = compositeBuffer();
+        CompositeByteBuf cbuf = newCompositeBuffer();
         cbuf.addComponent(EMPTY_BUFFER);
         cbuf.addComponent(EMPTY_BUFFER);
 
@@ -1185,7 +1240,7 @@ public abstract class AbstractCompositeByteBufTest extends AbstractByteBufTest {
 
     @Test
     public void testEmptyIterator() {
-        CompositeByteBuf cbuf = compositeBuffer();
+        CompositeByteBuf cbuf = newCompositeBuffer();
 
         Iterator<ByteBuf> it = cbuf.iterator();
         assertFalse(it.hasNext());
@@ -1199,33 +1254,43 @@ public abstract class AbstractCompositeByteBufTest extends AbstractByteBufTest {
         cbuf.release();
     }
 
-    @Test(expected = ConcurrentModificationException.class)
+    @Test
     public void testIteratorConcurrentModificationAdd() {
-        CompositeByteBuf cbuf = compositeBuffer();
+        CompositeByteBuf cbuf = newCompositeBuffer();
         cbuf.addComponent(EMPTY_BUFFER);
 
-        Iterator<ByteBuf> it = cbuf.iterator();
+        final Iterator<ByteBuf> it = cbuf.iterator();
         cbuf.addComponent(EMPTY_BUFFER);
 
         assertTrue(it.hasNext());
         try {
-            it.next();
+            assertThrows(ConcurrentModificationException.class, new Executable() {
+                @Override
+                public void execute() {
+                    it.next();
+                }
+            });
         } finally {
             cbuf.release();
         }
     }
 
-    @Test(expected = ConcurrentModificationException.class)
+    @Test
     public void testIteratorConcurrentModificationRemove() {
-        CompositeByteBuf cbuf = compositeBuffer();
+        CompositeByteBuf cbuf = newCompositeBuffer();
         cbuf.addComponent(EMPTY_BUFFER);
 
-        Iterator<ByteBuf> it = cbuf.iterator();
+        final Iterator<ByteBuf> it = cbuf.iterator();
         cbuf.removeComponent(0);
 
         assertTrue(it.hasNext());
         try {
-            it.next();
+            assertThrows(ConcurrentModificationException.class, new Executable() {
+                @Override
+                public void execute() {
+                    it.next();
+                }
+            });
         } finally {
             cbuf.release();
         }
@@ -1276,7 +1341,7 @@ public abstract class AbstractCompositeByteBufTest extends AbstractByteBufTest {
         ByteBuf s3 = s2.readRetainedSlice(2); // 4
         ByteBuf s4 = s3.readRetainedSlice(2); // 5
 
-        ByteBuf composite = Unpooled.compositeBuffer()
+        ByteBuf composite = newCompositeBuffer()
             .addComponent(s1)
             .addComponents(s2, s3, s4)
             .order(ByteOrder.LITTLE_ENDIAN);
@@ -1301,7 +1366,7 @@ public abstract class AbstractCompositeByteBufTest extends AbstractByteBufTest {
         ByteBuf b2 = Unpooled.buffer(2).writeShort(2);
 
         // composite takes ownership of s1 and s2
-        ByteBuf composite = Unpooled.compositeBuffer()
+        ByteBuf composite = newCompositeBuffer()
             .addComponents(b1, b2);
 
         assertEquals(4, composite.capacity());
@@ -1331,7 +1396,7 @@ public abstract class AbstractCompositeByteBufTest extends AbstractByteBufTest {
         ByteBuf b2 = b1.retainedSlice(b1.readerIndex(), 2);
 
         // composite takes ownership of b1 and b2
-        ByteBuf composite = Unpooled.compositeBuffer()
+        ByteBuf composite = newCompositeBuffer()
             .addComponents(b1, b2);
 
         assertEquals(4, composite.capacity());
@@ -1387,12 +1452,12 @@ public abstract class AbstractCompositeByteBufTest extends AbstractByteBufTest {
         testDecompose(310, 0, 0);
     }
 
-    private static void testDecompose(int offset, int length, int expectedListSize) {
+    private void testDecompose(int offset, int length, int expectedListSize) {
         byte[] bytes = new byte[1024];
         PlatformDependent.threadLocalRandom().nextBytes(bytes);
         ByteBuf buf = wrappedBuffer(bytes);
 
-        CompositeByteBuf composite = compositeBuffer();
+        CompositeByteBuf composite = newCompositeBuffer();
         composite.addComponents(true,
                                 buf.retainedSlice(100, 200),
                                 buf.retainedSlice(300, 400),
@@ -1412,6 +1477,64 @@ public abstract class AbstractCompositeByteBufTest extends AbstractByteBufTest {
         }
     }
 
+    @Test
+    public void testDecomposeReturnNonUnwrappedBuffer() {
+        ByteBuf buf = PooledByteBufAllocator.DEFAULT.buffer(1024);
+        buf.writeZero(1024);
+        ByteBuf sliced = buf.retainedSlice(100, 200);
+        sliced.retain();
+        assertEquals(2, sliced.refCnt());
+        CompositeByteBuf composite = newCompositeBuffer();
+        composite.addComponents(true, sliced);
+
+        List<ByteBuf> bufferList = composite.decompose(0, 100);
+        assertEquals(1, bufferList.size());
+        ByteBuf decomposed = bufferList.get(0);
+        assertSame(sliced.refCnt(), decomposed.refCnt());
+        decomposed.release();
+
+        assertSame(sliced.refCnt(), decomposed.refCnt());
+
+        composite.release();
+        buf.release();
+
+        for (ByteBuf buffer: bufferList) {
+            assertEquals(0, buffer.refCnt());
+        }
+    }
+
+    @Test
+    public void testDecomposeReturnNonUnwrappedBuffers() {
+        ByteBuf buf = PooledByteBufAllocator.DEFAULT.buffer(1024);
+        buf.writeZero(1024);
+        ByteBuf sliced = buf.retainedSlice(100, 200);
+        ByteBuf sliced2 = buf.retainedSlice(400, 100);
+        sliced.retain();
+        sliced2.retain();
+        assertEquals(2, sliced.refCnt());
+        CompositeByteBuf composite = compositeBuffer();
+        composite.addComponents(true, sliced);
+        composite.addComponents(true, sliced2);
+
+        List<ByteBuf> bufferList = composite.decompose(100, 150);
+        assertEquals(2, bufferList.size());
+        ByteBuf decomposed = bufferList.get(0);
+        ByteBuf decomposed2 = bufferList.get(1);
+        assertSame(sliced.refCnt(), decomposed.refCnt());
+        decomposed.release();
+        decomposed2.release();
+
+        assertSame(sliced.refCnt(), decomposed.refCnt());
+        assertSame(sliced2.refCnt(), decomposed2.refCnt());
+
+        composite.release();
+        buf.release();
+
+        for (ByteBuf buffer: bufferList) {
+            assertEquals(0, buffer.refCnt());
+        }
+    }
+
     @Test
     public void testComponentsLessThanLowerBound() {
         try {
@@ -1443,4 +1566,243 @@ public abstract class AbstractCompositeByteBufTest extends AbstractByteBufTest {
         assertTrue(buf.release());
     }
 
+    @Test
+    public void testDiscardSomeReadBytesCorrectlyUpdatesLastAccessed() {
+        testDiscardCorrectlyUpdatesLastAccessed(true);
+    }
+
+    @Test
+    public void testDiscardReadBytesCorrectlyUpdatesLastAccessed() {
+        testDiscardCorrectlyUpdatesLastAccessed(false);
+    }
+
+    private void testDiscardCorrectlyUpdatesLastAccessed(boolean discardSome) {
+        CompositeByteBuf cbuf = newCompositeBuffer();
+        List<ByteBuf> buffers = new ArrayList<ByteBuf>(4);
+        for (int i = 0; i < 4; i++) {
+            ByteBuf buf = buffer().writeInt(i);
+            cbuf.addComponent(true, buf);
+            buffers.add(buf);
+        }
+
+        // Skip the first 2 bytes which means even if we call discard*ReadBytes() later we can no drop the first
+        // component as it is still used.
+        cbuf.skipBytes(2);
+        if (discardSome) {
+            cbuf.discardSomeReadBytes();
+        } else {
+            cbuf.discardReadBytes();
+        }
+        assertEquals(4, cbuf.numComponents());
+
+        // Now skip 3 bytes which means we should be able to drop the first component on the next discard*ReadBytes()
+        // call.
+        cbuf.skipBytes(3);
+
+        if (discardSome) {
+            cbuf.discardSomeReadBytes();
+        } else {
+            cbuf.discardReadBytes();
+        }
+        assertEquals(3, cbuf.numComponents());
+        // Now skip again 3 bytes which should bring our readerIndex == start of the 3 component.
+        cbuf.skipBytes(3);
+
+        // Read one int (4 bytes) which should bring our readerIndex == start of the 4 component.
+        assertEquals(2, cbuf.readInt());
+        if (discardSome) {
+            cbuf.discardSomeReadBytes();
+        } else {
+            cbuf.discardReadBytes();
+        }
+
+        // Now all except the last component should have been dropped / released.
+        assertEquals(1, cbuf.numComponents());
+        assertEquals(3, cbuf.readInt());
+        if (discardSome) {
+            cbuf.discardSomeReadBytes();
+        } else {
+            cbuf.discardReadBytes();
+        }
+        assertEquals(0, cbuf.numComponents());
+
+        // These should have been released already.
+        for (ByteBuf buffer: buffers) {
+            assertEquals(0, buffer.refCnt());
+        }
+        assertTrue(cbuf.release());
+    }
+
+    // See https://github.com/netty/netty/issues/11612
+    @Test
+    public void testAddComponentWithNullEntry() {
+        final ByteBuf buffer = Unpooled.buffer(8).writeZero(8);
+        final CompositeByteBuf compositeByteBuf = compositeBuffer(Integer.MAX_VALUE);
+        try {
+            compositeByteBuf.addComponents(true, new ByteBuf[] { buffer, null });
+            assertEquals(8, compositeByteBuf.readableBytes());
+            assertEquals(1, compositeByteBuf.numComponents());
+        } finally {
+            compositeByteBuf.release();
+        }
+    }
+
+    @Test
+    public void testOverflowWhileAddingComponent() {
+        int capacity = 1024 * 1024; // 1MB
+        final ByteBuf buffer = Unpooled.buffer(capacity).writeZero(capacity);
+        final CompositeByteBuf compositeByteBuf = compositeBuffer(Integer.MAX_VALUE);
+
+        try {
+            assertThrows(IllegalArgumentException.class, new Executable() {
+                @Override
+                public void execute() {
+                    for (int i = 0; i >= 0; i += buffer.readableBytes()) {
+                        ByteBuf duplicate = buffer.duplicate();
+                        compositeByteBuf.addComponent(duplicate);
+                        duplicate.retain();
+                    }
+                }
+            });
+        } finally {
+            compositeByteBuf.release();
+        }
+    }
+
+    @Test
+    public void testOverflowWhileAddingComponentsViaVarargs() {
+        int capacity = 1024 * 1024; // 1MB
+        final ByteBuf buffer = Unpooled.buffer(capacity).writeZero(capacity);
+        final CompositeByteBuf compositeByteBuf = compositeBuffer(Integer.MAX_VALUE);
+
+        try {
+            assertThrows(IllegalArgumentException.class, new Executable() {
+                @Override
+                public void execute() {
+                    for (int i = 0; i >= 0; i += buffer.readableBytes()) {
+                        ByteBuf duplicate = buffer.duplicate();
+                        compositeByteBuf.addComponents(duplicate);
+                        duplicate.retain();
+                    }
+                }
+            });
+        } finally {
+            compositeByteBuf.release();
+        }
+    }
+
+    @Test
+    public void testOverflowWhileAddingComponentsViaIterable() {
+        int capacity = 1024 * 1024; // 1MB
+        final ByteBuf buffer = Unpooled.buffer(capacity).writeZero(capacity);
+        final CompositeByteBuf compositeByteBuf = compositeBuffer(Integer.MAX_VALUE);
+
+        try {
+            assertThrows(IllegalArgumentException.class, new Executable() {
+                @Override
+                public void execute() {
+                    for (int i = 0; i >= 0; i += buffer.readableBytes()) {
+                        ByteBuf duplicate = buffer.duplicate();
+                        compositeByteBuf.addComponents(Collections.singletonList(duplicate));
+                        duplicate.retain();
+                    }
+                }
+            });
+        } finally {
+            compositeByteBuf.release();
+        }
+    }
+
+    @Test
+    public void testOverflowWhileUseConstructorWithOffset() {
+        int capacity = 1024 * 1024; // 1MB
+        final ByteBuf buffer = Unpooled.buffer(capacity).writeZero(capacity);
+        final List<ByteBuf> buffers = new ArrayList<ByteBuf>();
+        for (long i = 0; i <= Integer.MAX_VALUE; i += capacity) {
+            buffers.add(buffer.duplicate());
+        }
+        // Add one more
+        buffers.add(buffer.duplicate());
+
+        try {
+            assertThrows(IllegalArgumentException.class, new Executable() {
+                @Override
+                public void execute() {
+                    ByteBuf[] bufferArray = buffers.toArray(new ByteBuf[0]);
+                    new CompositeByteBuf(ALLOC, false, Integer.MAX_VALUE, bufferArray, 0);
+                }
+            });
+        } finally {
+            buffer.release();
+        }
+    }
+
+    @Test
+    public void testNotOverflowWhileUseConstructorWithOffset() {
+        int capacity = 1024 * 1024; // 1MB
+        final ByteBuf buffer = Unpooled.buffer(capacity).writeZero(capacity);
+        final List<ByteBuf> buffers = new ArrayList<ByteBuf>();
+        for (long i = 0; i <= Integer.MAX_VALUE; i += capacity) {
+            buffers.add(buffer.duplicate());
+        }
+        // Add one more
+        buffers.add(buffer.duplicate());
+
+        ByteBuf[] bufferArray = buffers.toArray(new ByteBuf[0]);
+        CompositeByteBuf compositeByteBuf =
+                new CompositeByteBuf(ALLOC, false, Integer.MAX_VALUE, bufferArray, bufferArray.length - 1);
+        compositeByteBuf.release();
+    }
+
+    @Test
+    public void sliceOfCompositeBufferMustThrowISEAfterDiscardBytes() {
+        CompositeByteBuf composite = compositeBuffer();
+        composite.addComponent(true, buffer(8).writeZero(8));
+
+        ByteBuf slice = composite.retainedSlice();
+        composite.skipBytes(slice.readableBytes());
+        composite.discardSomeReadBytes();
+
+        try {
+            slice.readByte();
+            fail("Expected readByte of discarded slice to throw.");
+        } catch (IllegalStateException ignore) {
+            // Good.
+        } finally {
+            slice.release();
+            composite.release();
+        }
+    }
+
+    @Test
+    public void forEachByteOnNestedCompositeByteBufMustSeeEntireFlattenedContents() {
+        CompositeByteBuf buf = newCompositeBuffer();
+        buf.addComponent(true, newCompositeBuffer().addComponents(
+                true,
+                wrappedBuffer(new byte[] {1, 2, 3}),
+                wrappedBuffer(new byte[] {4, 5, 6})));
+        final byte[] arrayAsc = new byte[6];
+        final byte[] arrayDesc = new byte[6];
+        buf.forEachByte(new ByteProcessor() {
+            int index;
+
+            @Override
+            public boolean process(byte value) throws Exception {
+                arrayAsc[index++] = value;
+                return true;
+            }
+        });
+        buf.forEachByteDesc(new ByteProcessor() {
+            int index;
+
+            @Override
+            public boolean process(byte value) throws Exception {
+                arrayDesc[index++] = value;
+                return true;
+            }
+        });
+        assertArrayEquals(new byte[] {1, 2, 3, 4, 5, 6}, arrayAsc);
+        assertArrayEquals(new byte[] {6, 5, 4, 3, 2, 1}, arrayDesc);
+        buf.release();
+    }
 }
