@@ -5,7 +5,7 @@
  * version 2.0 (the "License"); you may not use this file except in compliance
  * with the License. You may obtain a copy of the License at:
  *
- *   http://www.apache.org/licenses/LICENSE-2.0
+ *   https://www.apache.org/licenses/LICENSE-2.0
  *
  * Unless required by applicable law or agreed to in writing, software
  * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
@@ -15,26 +15,79 @@
  */
 package io.netty.handler.codec.http;
 
-import java.net.InetAddress;
-import java.net.InetSocketAddress;
-import java.nio.charset.StandardCharsets;
-import java.util.ArrayList;
-import java.util.Collections;
-import java.util.List;
-
+import io.netty.util.AsciiString;
 import io.netty.util.CharsetUtil;
 import io.netty.util.ReferenceCountUtil;
-import org.junit.Test;
+import org.junit.jupiter.api.Test;
+import org.junit.jupiter.api.function.Executable;
+import org.junit.jupiter.params.ParameterizedTest;
+import org.junit.jupiter.params.provider.MethodSource;
+import org.junit.jupiter.params.provider.ValueSource;
+
+import java.net.InetAddress;
+import java.net.InetSocketAddress;
+import java.net.URI;
+import java.nio.charset.StandardCharsets;
+import java.util.ArrayList;
+import java.util.List;
 
 import static io.netty.handler.codec.http.HttpHeadersTestUtils.of;
-import static org.junit.Assert.assertEquals;
-import static org.junit.Assert.assertFalse;
-import static org.junit.Assert.assertNull;
-import static org.junit.Assert.assertTrue;
-import static org.junit.Assert.fail;
+import static io.netty.handler.codec.http.HttpUtil.normalizeAndGetContentLength;
+import static io.netty.handler.codec.http.HttpUtil.validateToken;
+import static java.util.Collections.singletonList;
+import static org.junit.jupiter.api.Assertions.assertEquals;
+import static org.junit.jupiter.api.Assertions.assertFalse;
+import static org.junit.jupiter.api.Assertions.assertNull;
+import static org.junit.jupiter.api.Assertions.assertThrows;
+import static org.junit.jupiter.api.Assertions.assertTrue;
+import static org.junit.jupiter.api.Assertions.fail;
 
 public class HttpUtilTest {
 
+    @Test
+    public void testRecognizesOriginForm() {
+        // Origin form: https://tools.ietf.org/html/rfc7230#section-5.3.1
+        assertTrue(HttpUtil.isOriginForm(URI.create("/where?q=now")));
+        // Absolute form: https://tools.ietf.org/html/rfc7230#section-5.3.2
+        assertFalse(HttpUtil.isOriginForm(URI.create("http://www.example.org/pub/WWW/TheProject.html")));
+        // Authority form: https://tools.ietf.org/html/rfc7230#section-5.3.3
+        assertFalse(HttpUtil.isOriginForm(URI.create("www.example.com:80")));
+        // Asterisk form: https://tools.ietf.org/html/rfc7230#section-5.3.4
+        assertFalse(HttpUtil.isOriginForm(URI.create("*")));
+    }
+
+    @Test
+    public void testRecognizesAsteriskForm() {
+        // Asterisk form: https://tools.ietf.org/html/rfc7230#section-5.3.4
+        assertTrue(HttpUtil.isAsteriskForm(URI.create("*")));
+        // Origin form: https://tools.ietf.org/html/rfc7230#section-5.3.1
+        assertFalse(HttpUtil.isAsteriskForm(URI.create("/where?q=now")));
+        // Absolute form: https://tools.ietf.org/html/rfc7230#section-5.3.2
+        assertFalse(HttpUtil.isAsteriskForm(URI.create("http://www.example.org/pub/WWW/TheProject.html")));
+        // Authority form: https://tools.ietf.org/html/rfc7230#section-5.3.3
+        assertFalse(HttpUtil.isAsteriskForm(URI.create("www.example.com:80")));
+    }
+
+    @ParameterizedTest
+    @ValueSource(strings = {
+            "http://localhost/\r\n",
+            "/r\r\n?q=1",
+            "http://localhost/\r\n?q=1",
+            "/r\r\n/?q=1",
+            "http://localhost/\r\n/?q=1",
+            "/r\r\n",
+            "http://localhost/ HTTP/1.1\r\n\r\nPOST /p HTTP/1.1\r\n\r\n",
+            "/r HTTP/1.1\r\n\r\nPOST /p HTTP/1.1\r\n\r\n",
+            "GET ",
+            " GET",
+            "HTTP/ 1.1",
+            "HTTP/\r0.9",
+            "HTTP/\n1.1",
+    })
+    public void requestLineTokenValidationMustRejectInvalidTokens(String token) throws Exception {
+        assertFalse(HttpUtil.isEncodingSafeStartLineToken(token));
+    }
+
     @Test
     public void testRemoveTransferEncodingIgnoreCase() {
         HttpMessage message = new DefaultHttpResponse(HttpVersion.HTTP_1_1, HttpResponseStatus.OK);
@@ -76,19 +129,66 @@ public class HttpUtilTest {
 
     @Test
     public void testGetCharset() {
-        String NORMAL_CONTENT_TYPE = "text/html; charset=utf-8";
-        String UPPER_CASE_NORMAL_CONTENT_TYPE = "TEXT/HTML; CHARSET=UTF-8";
+        testGetCharsetUtf8("text/html; charset=utf-8");
+    }
+
+    @Test
+    public void testGetCharsetNoSpace() {
+        testGetCharsetUtf8("text/html;charset=utf-8");
+    }
+
+    @Test
+    public void testGetCharsetQuoted() {
+        testGetCharsetUtf8("text/html; charset=\"utf-8\"");
+    }
+
+    @Test
+    public void testGetCharsetNoSpaceQuoted() {
+        testGetCharsetUtf8("text/html;charset=\"utf-8\"");
+    }
+
+    private void testGetCharsetUtf8(String contentType) {
+        String UPPER_CASE_NORMAL_CONTENT_TYPE = contentType.toUpperCase();
 
         HttpMessage message = new DefaultHttpResponse(HttpVersion.HTTP_1_1, HttpResponseStatus.OK);
-        message.headers().set(HttpHeaderNames.CONTENT_TYPE, NORMAL_CONTENT_TYPE);
+        message.headers().set(HttpHeaderNames.CONTENT_TYPE, contentType);
         assertEquals(CharsetUtil.UTF_8, HttpUtil.getCharset(message));
-        assertEquals(CharsetUtil.UTF_8, HttpUtil.getCharset(NORMAL_CONTENT_TYPE));
+        assertEquals(CharsetUtil.UTF_8, HttpUtil.getCharset(contentType));
 
         message.headers().set(HttpHeaderNames.CONTENT_TYPE, UPPER_CASE_NORMAL_CONTENT_TYPE);
         assertEquals(CharsetUtil.UTF_8, HttpUtil.getCharset(message));
         assertEquals(CharsetUtil.UTF_8, HttpUtil.getCharset(UPPER_CASE_NORMAL_CONTENT_TYPE));
     }
 
+    @Test
+    public void testGetCharsetNoLeadingQuotes() {
+        testGetCharsetInvalidQuotes("text/html;charset=utf-8\"");
+    }
+
+    @Test
+    public void testGetCharsetNoTrailingQuotes() {
+        testGetCharsetInvalidQuotes("text/html;charset=\"utf-8");
+    }
+
+    @Test
+    public void testGetCharsetOnlyQuotes() {
+        testGetCharsetInvalidQuotes("text/html;charset=\"\"");
+    }
+
+    private static void testGetCharsetInvalidQuotes(String contentType) {
+        String UPPER_CASE_NORMAL_CONTENT_TYPE = contentType.toUpperCase();
+
+        HttpMessage message = new DefaultHttpResponse(HttpVersion.HTTP_1_1, HttpResponseStatus.OK);
+        message.headers().set(HttpHeaderNames.CONTENT_TYPE, contentType);
+        assertEquals(CharsetUtil.ISO_8859_1, HttpUtil.getCharset(message, CharsetUtil.ISO_8859_1));
+        assertEquals(CharsetUtil.ISO_8859_1, HttpUtil.getCharset(contentType, CharsetUtil.ISO_8859_1));
+
+        message.headers().set(HttpHeaderNames.CONTENT_TYPE, UPPER_CASE_NORMAL_CONTENT_TYPE);
+        assertEquals(CharsetUtil.ISO_8859_1, HttpUtil.getCharset(message, CharsetUtil.ISO_8859_1));
+        assertEquals(CharsetUtil.ISO_8859_1, HttpUtil.getCharset(UPPER_CASE_NORMAL_CONTENT_TYPE,
+                CharsetUtil.ISO_8859_1));
+    }
+
     @Test
     public void testGetCharsetIfNotLastParameter() {
         String NORMAL_CONTENT_TYPE_WITH_PARAMETERS = "application/soap-xml; charset=utf-8; "
@@ -109,6 +209,7 @@ public class HttpUtilTest {
     public void testGetCharset_defaultValue() {
         final String SIMPLE_CONTENT_TYPE = "text/html";
         final String CONTENT_TYPE_WITH_INCORRECT_CHARSET = "text/html; charset=UTFFF";
+        final String CONTENT_TYPE_WITH_ILLEGAL_CHARSET_NAME = "text/html; charset=!illegal!";
 
         HttpMessage message = new DefaultHttpResponse(HttpVersion.HTTP_1_1, HttpResponseStatus.OK);
         message.headers().set(HttpHeaderNames.CONTENT_TYPE, SIMPLE_CONTENT_TYPE);
@@ -127,6 +228,15 @@ public class HttpUtilTest {
         assertEquals(CharsetUtil.UTF_8, HttpUtil.getCharset(message, StandardCharsets.UTF_8));
         assertEquals(CharsetUtil.UTF_8,
                      HttpUtil.getCharset(CONTENT_TYPE_WITH_INCORRECT_CHARSET, StandardCharsets.UTF_8));
+
+        message.headers().set(HttpHeaderNames.CONTENT_TYPE, CONTENT_TYPE_WITH_ILLEGAL_CHARSET_NAME);
+        assertEquals(CharsetUtil.ISO_8859_1, HttpUtil.getCharset(message));
+        assertEquals(CharsetUtil.ISO_8859_1, HttpUtil.getCharset(CONTENT_TYPE_WITH_ILLEGAL_CHARSET_NAME));
+
+        message.headers().set(HttpHeaderNames.CONTENT_TYPE, CONTENT_TYPE_WITH_ILLEGAL_CHARSET_NAME);
+        assertEquals(CharsetUtil.UTF_8, HttpUtil.getCharset(message, StandardCharsets.UTF_8));
+        assertEquals(CharsetUtil.UTF_8,
+                HttpUtil.getCharset(CONTENT_TYPE_WITH_ILLEGAL_CHARSET_NAME, StandardCharsets.UTF_8));
     }
 
     @Test
@@ -189,7 +299,7 @@ public class HttpUtilTest {
         HttpMessage message = new DefaultHttpResponse(HttpVersion.HTTP_1_1, HttpResponseStatus.OK);
         message.headers().add(HttpHeaderNames.TRANSFER_ENCODING, "chunked");
         HttpUtil.setTransferEncodingChunked(message, true);
-        List<String> expected = Collections.singletonList("chunked");
+        List<String> expected = singletonList("chunked");
         assertEquals(expected, message.headers().getAll(HttpHeaderNames.TRANSFER_ENCODING));
     }
 
@@ -295,6 +405,12 @@ public class HttpUtilTest {
         assertEquals("[::1]", HttpUtil.formatHostnameForHttp(socketAddress));
     }
 
+    @Test
+    public void testIpv6UnresolvedExplicitAddress()  {
+        InetSocketAddress socketAddress = InetSocketAddress.createUnresolved("[2001:4860:4860:0:0:0:0:8888]", 8080);
+        assertEquals("[2001:4860:4860:0:0:0:0:8888]", HttpUtil.formatHostnameForHttp(socketAddress));
+    }
+
     @Test
     public void testIpv4() throws Exception  {
         InetSocketAddress socketAddress = new InetSocketAddress(InetAddress.getByName("10.0.0.1"), 8080);
@@ -338,4 +454,85 @@ public class HttpUtilTest {
                 HttpHeaderNames.CONNECTION, HttpHeaderValues.UPGRADE + ", " + HttpHeaderValues.KEEP_ALIVE);
         assertTrue(HttpUtil.isKeepAlive(http11Message));
     }
+
+    @Test
+    public void normalizeAndGetContentLengthEmpty() {
+        testNormalizeAndGetContentLengthInvalidContentLength("");
+    }
+
+    @Test
+    public void normalizeAndGetContentLengthNotANumber() {
+        testNormalizeAndGetContentLengthInvalidContentLength("foo");
+    }
+
+    @Test
+    public void normalizeAndGetContentLengthNegative() {
+        testNormalizeAndGetContentLengthInvalidContentLength("-1");
+    }
+
+    private static void testNormalizeAndGetContentLengthInvalidContentLength(final String contentLengthField) {
+        assertThrows(IllegalArgumentException.class, new Executable() {
+            @Override
+            public void execute() {
+                normalizeAndGetContentLength(singletonList(contentLengthField), false, false);
+            }
+        });
+    }
+
+    private static List<Character> validTokenChars() {
+        List<Character> list = new ArrayList<Character>();
+        for (char c = '0'; c <= '9'; c++) {
+            list.add(c);
+        }
+        for (char c = 'a'; c <= 'z'; c++) {
+            list.add(c);
+        }
+        for (char c = 'A'; c <= 'Z'; c++) {
+            list.add(c);
+        }
+
+        // Unreserved characters:
+        list.add('-');
+        list.add('.');
+        list.add('_');
+        list.add('~');
+
+        // Token special characters:
+        list.add('!');
+        list.add('#');
+        list.add('$');
+        list.add('%');
+        list.add('&');
+        list.add('\'');
+        list.add('*');
+        list.add('+');
+        list.add('^');
+        list.add('`');
+        list.add('|');
+
+        return list;
+    }
+
+    @ParameterizedTest
+    @MethodSource("validTokenChars")
+    public void testValidTokenChars(char validChar) {
+        AsciiString asciiStringToken =
+                new AsciiString(new byte[] { 'G', 'E', (byte) validChar, 'T' });
+        String token = "GE" + validChar + 'T';
+        assertEquals(-1, validateToken(asciiStringToken));
+        assertEquals(-1, validateToken(token));
+    }
+
+    @ParameterizedTest
+    @ValueSource(chars = {
+            '(', ')', ',', '/', ':', ';', '<', '=', '>', '?', '@',
+            '\"', '[', '\\', ']', '{', '}', '\u0000', ' ', '\u007f', 'ÿ'
+    })
+    public void testInvalidTokenChars(char invalidChar) {
+        AsciiString asciiStringToken =
+                new AsciiString(new byte[] { 'G', 'E', (byte) invalidChar, 'T' });
+        String token = "GE" + invalidChar + 'T';
+        assertEquals(2, validateToken(asciiStringToken));
+        assertEquals(2, validateToken(token));
+    }
 }
