@@ -5,7 +5,7 @@
  * version 2.0 (the "License"); you may not use this file except in compliance
  * with the License. You may obtain a copy of the License at:
  *
- *   http://www.apache.org/licenses/LICENSE-2.0
+ *   https://www.apache.org/licenses/LICENSE-2.0
  *
  * Unless required by applicable law or agreed to in writing, software
  * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
@@ -13,19 +13,22 @@
  * License for the specific language governing permissions and limitations
  * under the License.
  */
-
 package io.netty.handler.ssl;
 
-import org.junit.Assume;
+import org.junit.jupiter.api.BeforeAll;
 
 import javax.net.ssl.SSLException;
 import java.io.File;
 
 public class OpenSslServerContextTest extends SslContextTest {
 
+    @BeforeAll
+    public static void checkOpenSsl() {
+        OpenSsl.ensureAvailability();
+    }
+
     @Override
-    protected SslContext newServerContext(File crtFile, File keyFile, String pass) throws SSLException {
-        Assume.assumeTrue(OpenSsl.isAvailable());
+    protected SslContext newSslContext(File crtFile, File keyFile, String pass) throws SSLException {
         return new OpenSslServerContext(crtFile, keyFile, pass);
     }
 }
